@@ -1,5 +1,5 @@
-/* 
- * Copyright (C) 2007 OpenedHand Ltd.
+/*
+ * Copyright (C) 2007, 2008 OpenedHand Ltd.
  *
  * Author: Jorn Baayen <jorn@openedhand.com>
  *
@@ -28,18 +28,20 @@
  */
 
 #include <gobject/gvaluecollector.h>
+#include <gmodule.h>
 #include <libsoup/soup-date.h>
 #include <uuid/uuid.h>
 #include <string.h>
 
 #include "gupnp-service.h"
-#include "gupnp-service-private.h"
+#include "gupnp-root-device.h"
 #include "gupnp-context-private.h"
 #include "gupnp-marshal.h"
 #include "gupnp-error.h"
-#include "accept-language.h"
+#include "http-headers.h"
 #include "gena-protocol.h"
 #include "xml-util.h"
+#include "gvalue-util.h"
 
 G_DEFINE_TYPE (GUPnPService,
                gupnp_service,
@@ -57,6 +59,9 @@ struct _GUPnPServicePrivate {
         GQueue          *notify_queue;
 
         gboolean         notify_frozen;
+
+        GList           *pending_messages; /* Pending SoupMessages from this
+                                              service */
 };
 
 enum {
@@ -73,7 +78,7 @@ enum {
 
 static guint signals[LAST_SIGNAL];
 
-static xmlChar *
+static char *
 create_property_set (GQueue *queue);
 
 static void
@@ -130,7 +135,7 @@ struct _GUPnPServiceAction {
 
         xmlNode     *node;
 
-        xmlNode     *response_node;
+        GString     *response_str;
 };
 
 /**
@@ -151,7 +156,7 @@ gupnp_service_action_get_name (GUPnPServiceAction *action)
  * gupnp_service_action_get_locales
  * @action: A #GUPnPServiceAction
  *
- * Return value: An ordered (preferred first) #GList of locales preferred by 
+ * Return value: An ordered (preferred first) #GList of locales preferred by
  * the client. Free list and elements after use.
  **/
 GList *
@@ -220,8 +225,6 @@ gupnp_service_action_get_valist (GUPnPServiceAction *action,
                         g_warning ("Error lcopying value: %s\n", copy_error);
 
                         g_free (copy_error);
-
-                        return;
                 }
 
                 arg_name = va_arg (var_args, const char *);
@@ -252,7 +255,7 @@ gupnp_service_action_get_value (GUPnPServiceAction *action,
                 if (strcmp ((char *) node->name, argument) != 0)
                         continue;
 
-                xml_util_node_get_content_value (node, value);
+                gvalue_util_set_value_from_xml_node (value, node);
 
                 break;
         }
@@ -296,7 +299,7 @@ gupnp_service_action_set_valist (GUPnPServiceAction *action,
         GType arg_type;
         GValue value = {0, };
         char *collect_error;
-        
+
         g_return_if_fail (action != NULL);
 
         collect_error = NULL;
@@ -307,21 +310,19 @@ gupnp_service_action_set_valist (GUPnPServiceAction *action,
                 g_value_init (&value, arg_type);
 
                 G_VALUE_COLLECT (&value, var_args, 0, &collect_error);
-                if (collect_error) {
+                if (!collect_error) {
+                        gupnp_service_action_set_value (action,
+                                                        arg_name, &value);
+
+                        g_value_unset (&value);
+
+                } else {
                         g_warning ("Error collecting value: %s\n",
                                    collect_error);
 
                         g_free (collect_error);
-
-                        g_value_unset (&value);
-
-                        return;
                 }
 
-                gupnp_service_action_set_value (action, arg_name, &value);
-
-                g_value_unset (&value);
-
                 arg_name = va_arg (var_args, const char *);
         }
 }
@@ -339,13 +340,11 @@ gupnp_service_action_set_value (GUPnPServiceAction *action,
                                 const char         *argument,
                                 const GValue       *value)
 {
-        GValue transformed_value = {0, };
-
         g_return_if_fail (action != NULL);
         g_return_if_fail (argument != NULL);
         g_return_if_fail (value != NULL);
 
-        if (action->msg->status == SOUP_STATUS_INTERNAL_SERVER_ERROR) {
+        if (action->msg->status_code == SOUP_STATUS_INTERNAL_SERVER_ERROR) {
                 g_warning ("Calling gupnp_service_action_set_value() after "
                            "having called gupnp_service_action_return_error() "
                            "is not allowed.");
@@ -353,22 +352,10 @@ gupnp_service_action_set_value (GUPnPServiceAction *action,
                 return;
         }
 
-        /* Transform to string */
-        g_value_init (&transformed_value, G_TYPE_STRING);
-        if (g_value_transform (value, &transformed_value)) {
-                /* Append to response */
-                xmlNewTextChild (action->response_node,
-                                 NULL,
-                                 (const xmlChar *) argument,
-                                 (xmlChar *) g_value_get_string
-                                                    (&transformed_value));
-        } else {
-                g_warning ("Failed to transform value of type %s to "
-                           "string.", G_VALUE_TYPE_NAME (value));
-        }
-
-        /* Cleanup */
-        g_value_unset (&transformed_value);
+        /* Append to response */
+        xml_util_start_element (action->response_str, argument);
+        gvalue_util_value_append_to_xml_string (value, action->response_str);
+        xml_util_end_element (action->response_str, argument);
 }
 
 /**
@@ -402,10 +389,6 @@ gupnp_service_action_return_error (GUPnPServiceAction *action,
                                    guint               error_code,
                                    const char         *error_description)
 {
-        xmlNode *node;
-        xmlNs *ns;
-        char *tmp;
-
         g_return_if_fail (action != NULL);
 
         switch (error_code) {
@@ -434,47 +417,37 @@ gupnp_service_action_return_error (GUPnPServiceAction *action,
                 break;
         }
 
-        /* Replace response_node with a SOAP Fault */
-        xmlFreeNode (action->response_node);
-        action->response_node = xmlNewNode (NULL,
-                                            (const xmlChar *) "Fault");
+        /* Replace response_str with a SOAP Fault */
+        g_string_erase (action->response_str, 0, -1);
 
-        xmlNewTextChild (action->response_node,
-                         NULL,
-                         (const xmlChar *) "faultcode",
-                         (const xmlChar *) "s:Client");
+        xml_util_start_element (action->response_str, "s:Fault");
 
-        xmlNewTextChild (action->response_node,
-                         NULL,
-                         (const xmlChar *) "faultstring",
-                         (const xmlChar *) "UPnPError");
+        xml_util_start_element (action->response_str, "faultcode");
+        g_string_append (action->response_str, "s:Client");
+        xml_util_end_element (action->response_str, "faultcode");
 
-        node = xmlNewTextChild (action->response_node,
-                                NULL,
-                                (const xmlChar *) "detail",
-                                NULL);
+        xml_util_start_element (action->response_str, "faultstring");
+        g_string_append (action->response_str, "UPnPError");
+        xml_util_end_element (action->response_str, "faultstring");
 
-        node = xmlNewTextChild (node,
-                                NULL,
-                                (const xmlChar *) "UPnPError",
-                                NULL);
+        xml_util_start_element (action->response_str, "detail");
 
-        ns = xmlNewNs (node,
-                       (const xmlChar *) "urn:schemas-upnp-org:control-1-0",
-                       (const xmlChar *) "u");
-        xmlSetNs (node, ns);
+        xml_util_start_element (action->response_str,
+                                "UPnPError "
+                                "xmlns=\"urn:schemas-upnp-org:control-1-0\"");
 
-        tmp = g_strdup_printf ("%u", error_code);
-        xmlNewTextChild (node,
-                         NULL,
-                         (const xmlChar *) "errorCode",
-                         (const xmlChar *) tmp);
-        g_free (tmp);
+        xml_util_start_element (action->response_str, "errorCode");
+        g_string_append_printf (action->response_str, "%u", error_code);
+        xml_util_end_element (action->response_str, "errorCode");
 
-        xmlNewTextChild (node,
-                         NULL,
-                         (const xmlChar *) "errorDescription",
-                         (const xmlChar *) error_description);
+        xml_util_start_element (action->response_str, "errorDescription");
+        xml_util_add_content (action->response_str, error_description);
+        xml_util_end_element (action->response_str, "errorDescription");
+
+        xml_util_end_element (action->response_str, "UPnPError");
+        xml_util_end_element (action->response_str, "detail");
+
+        xml_util_end_element (action->response_str, "s:Fault");
 
         soup_message_set_status (action->msg,
                                  SOUP_STATUS_INTERNAL_SERVER_ERROR);
@@ -497,31 +470,33 @@ gupnp_service_init (GUPnPService *device)
 }
 
 /* Generate a new action response node for @action_name */
-static xmlNode *
-new_action_response_node (GUPnPService *service,
-                          const char   *action_name)
+static GString *
+new_action_response_str (GUPnPService *service,
+                         const char   *action_name)
 {
-        xmlNode *node;
+        GString *str;
         const char *type;
-        char *tmp;
 
-        tmp = g_strdup_printf ("%sResponse", action_name);
-        node = xmlNewNode (NULL, (const xmlChar *) tmp);
-        g_free (tmp);
+        str = xml_util_new_string ();
+
+        g_string_append (str, "<u:");
+        g_string_append (str, action_name);
+        g_string_append (str, "Response xmlns:u=");
 
         type = gupnp_service_info_get_service_type
                                         (GUPNP_SERVICE_INFO (service));
         if (type != NULL) {
-                xmlNs *ns;
-
-                ns = xmlNewNs (node, (xmlChar *) type, (const xmlChar *) "u");
-                xmlSetNs (node, ns);
+                g_string_append_c (str, '"');
+                g_string_append (str, type);
+                g_string_append_c (str, '"');
         } else {
                 g_warning ("No serviceType defined. Control may not work "
                            "correctly.");
         }
 
-        return node;
+        g_string_append_c (str, '>');
+
+        return str;
 }
 
 /* Handle QueryStateVariable action */
@@ -589,14 +564,11 @@ control_server_handler (SoupServerContext *server_context,
         GUPnPService *service;
         GUPnPContext *context;
         GSSDPClient *client;
-        xmlDoc *doc, *response_doc;
-        xmlNode *action_node, *response_node;
-        xmlNs *ns;
+        xmlDoc *doc;
+        xmlNode *action_node;
         const char *soap_action, *action_name;
         char *end, *date;
         GUPnPServiceAction *action;
-        xmlChar *mem;
-        int size;
 
         service = GUPNP_SERVICE (user_data);
 
@@ -641,10 +613,10 @@ control_server_handler (SoupServerContext *server_context,
         /* Create action structure */
         action = g_slice_new (GUPnPServiceAction);
 
-        action->name          = action_name;
-        action->msg           = msg;
-        action->node          = action_node;
-        action->response_node = new_action_response_node (service, action_name);
+        action->name         = action_name;
+        action->msg          = msg;
+        action->node         = action_node;
+        action->response_str = new_action_response_str (service, action_name);
 
         /* QueryStateVariable? */
         if (strcmp (action_name, "QueryStateVariable") == 0)
@@ -665,41 +637,30 @@ control_server_handler (SoupServerContext *server_context,
                 }
         }
 
-        /* Embed action->response_node in a SOUP document */
-        response_doc = xmlNewDoc ((const xmlChar *) "1.0");
-        response_node = xmlNewDocNode (response_doc,
-                                       NULL,
-                                       (const xmlChar *) "Envelope",
-                                       NULL);
-        xmlDocSetRootElement (response_doc, response_node);
+        /* Embed action->response_str in a SOAP document */
+        g_string_prepend (action->response_str, 
+                          "<?xml version=\"1.0\"?>"
+                          "<s:Envelope xmlns:s="
+                                "\"http://schemas.xmlsoap.org/soap/envelope/\" "
+                          "s:encodingStyle="
+                                "\"http://schemas.xmlsoap.org/soap/encoding/\">"
+                          "<s:Body>");
 
-        ns = xmlNewNs (response_node,
-                       (const xmlChar *)
-                               "http://schemas.xmlsoap.org/soap/envelope/",
-                       (const xmlChar *) "s");
-        xmlSetNs (response_node, ns);
+        if (action->msg->status_code != SOUP_STATUS_INTERNAL_SERVER_ERROR) {
+                g_string_append (action->response_str, "</u:");
+                g_string_append (action->response_str, action_name);
+                g_string_append (action->response_str, "Response>");
+        }
 
-        response_node = xmlNewChild (response_node,
-                                     ns,
-                                     (const xmlChar *) "Body",
-                                     NULL);
-
-        if (msg->status_code == SOUP_STATUS_INTERNAL_SERVER_ERROR)
-                xmlSetNs (action->response_node, ns);
-
-        xmlAddChild (response_node, action->response_node);
-
-        /* Dump document to response */
-        xmlDocDumpMemory (response_doc, &mem, &size);
+        g_string_append (action->response_str,
+                         "</s:Body>"
+                         "</s:Envelope>");
 
         msg->response.owner  = SOUP_BUFFER_SYSTEM_OWNED;
-        msg->response.body   = g_strdup ((char *) mem);
-        msg->response.length = size;
+        msg->response.body   = g_string_free (action->response_str, FALSE);
+        msg->response.length = strlen (msg->response.body);
 
         /* Cleanup */
-        xmlFreeDoc (response_doc);
-        xmlFree (mem);
-
         g_slice_free (GUPnPServiceAction, action);
 
         /* Server header on response */
@@ -787,10 +748,10 @@ subscription_timeout (gpointer user_data)
         return FALSE;
 }
 
-/* Parse timeout header and return its value in seconds, or -1
+/* Parse timeout header and return its value in seconds, or 0
  * if infinite */
 static int
-parse_timeout (const char *timeout)
+parse_and_limit_timeout (const char *timeout)
 {
         int timeout_seconds;
 
@@ -798,14 +759,11 @@ parse_timeout (const char *timeout)
 
         if (strncmp (timeout, "Second-", strlen ("Second-")) == 0) {
                 /* We have a finite timeout */
-                timeout_seconds = atoi (timeout + strlen ("Second-"));
-                if (timeout_seconds < GENA_MIN_TIMEOUT) {
-                        g_warning ("Specified timeout is too short. Assuming "
-                                   "default of %d.", GENA_DEFAULT_TIMEOUT);
-
-                        timeout_seconds = GENA_DEFAULT_TIMEOUT;
-                }
-        }
+                timeout_seconds = CLAMP (atoi (timeout + strlen ("Second-")),
+                                         GENA_MIN_TIMEOUT,
+                                         GENA_MAX_TIMEOUT);
+        } else
+                timeout_seconds = GENA_MAX_TIMEOUT;
 
         return timeout_seconds;
 }
@@ -821,7 +779,7 @@ subscribe (GUPnPService *service,
         char *start, *end, *uri;
         int timeout_seconds;
         GQueue *queue;
-        xmlChar *mem;
+        char *mem;
         GList *l;
 
         data = g_slice_new0 (SubscriptionData);
@@ -858,7 +816,7 @@ subscribe (GUPnPService *service,
         data->sid     = generate_sid ();
 
         /* Add timeout */
-        timeout_seconds = parse_timeout (timeout);
+        timeout_seconds = parse_and_limit_timeout (timeout);
         if (timeout_seconds > 0) {
                 data->timeout_id = g_timeout_add (timeout_seconds * 1000,
                                                   subscription_timeout,
@@ -876,7 +834,7 @@ subscribe (GUPnPService *service,
         /* Send initial event message */
         queue = g_queue_new ();
 
-        for (l = service->priv->state_variables; l; l = l->next) {      
+        for (l = service->priv->state_variables; l; l = l->next) {
                 NotifyData *ndata;
 
                 ndata = g_slice_new0 (NotifyData);
@@ -899,12 +857,12 @@ subscribe (GUPnPService *service,
         }
 
         mem = create_property_set (queue);
-        notify_subscriber (data->sid, data, mem); 
+        notify_subscriber (data->sid, data, mem);
 
         /* Cleanup */
         g_queue_free (queue);
 
-        xmlFree (mem);
+        g_free (mem);
 }
 
 /* Resubscription request */
@@ -930,7 +888,7 @@ resubscribe (GUPnPService *service,
                 data->timeout_id = 0;
         }
 
-        timeout_seconds = parse_timeout (timeout);
+        timeout_seconds = parse_and_limit_timeout (timeout);
         if (timeout_seconds > 0) {
                 data->timeout_id = g_timeout_add (timeout_seconds * 1000,
                                                   subscription_timeout,
@@ -1057,13 +1015,20 @@ gupnp_service_constructor (GType                  type,
         introspection = gupnp_service_info_get_introspection (info, &error);
         if (introspection) {
                 state_variables =
-                        gupnp_service_introspection_list_state_variable_names
+                        gupnp_service_introspection_list_state_variables
                                                                 (introspection);
 
                 for (l = state_variables; l; l = l->next) {
+                        GUPnPServiceStateVariableInfo *variable;
+
+                        variable = l->data;
+
+                        if (!variable->send_events)
+                                continue;
+
                         service->priv->state_variables =
                                 g_list_prepend (service->priv->state_variables,
-                                                g_strdup (l->data));
+                                                g_strdup (variable->name));
                 }
 
                 g_object_unref (introspection);
@@ -1077,7 +1042,7 @@ gupnp_service_constructor (GType                  type,
 
         /* Get server */
         context = gupnp_service_info_get_context (info);
-        server = _gupnp_context_get_server (context);
+        server = gupnp_context_get_server (context);
 
         /* Run listener on controlURL */
         url = gupnp_service_info_get_control_url (info);
@@ -1187,6 +1152,8 @@ gupnp_service_dispose (GObject *object)
 {
         GUPnPService *service;
         GObjectClass *object_class;
+        GUPnPContext *context;
+        SoupSession *session;
 
         service = GUPNP_SERVICE (object);
 
@@ -1198,11 +1165,31 @@ gupnp_service_dispose (GObject *object)
                                 (service->priv->root_device,
                                  service->priv->notify_available_id);
                 }
-                
+
                 g_object_unref (service->priv->root_device);
                 service->priv->root_device = NULL;
         }
 
+        /* Cancel pending messages */
+        context = gupnp_service_info_get_context (GUPNP_SERVICE_INFO (service));
+        if (context)
+                session = _gupnp_context_get_session (context);
+        else
+                session = NULL; /* Not the first time dispose is called. */
+
+        while (service->priv->pending_messages) {
+                SoupMessage *msg;
+
+                msg = service->priv->pending_messages->data;
+
+                soup_message_set_status (msg, SOUP_STATUS_CANCELLED);
+                soup_session_cancel_message (session, msg);
+
+                service->priv->pending_messages =
+                        g_list_delete_link (service->priv->pending_messages,
+                                            service->priv->pending_messages);
+        }
+
         /* Call super */
         object_class = G_OBJECT_CLASS (gupnp_service_parent_class);
         object_class->dispose (object);
@@ -1254,7 +1241,7 @@ gupnp_service_class_init (GUPnPServiceClass *klass)
         object_class->finalize     = gupnp_service_finalize;
 
         info_class = GUPNP_SERVICE_INFO_CLASS (klass);
-        
+
         g_type_class_add_private (klass, sizeof (GUPnPServicePrivate));
 
         /**
@@ -1392,21 +1379,18 @@ gupnp_service_notify_valist (GUPnPService *service,
                 g_value_init (&value, var_type);
 
                 G_VALUE_COLLECT (&value, var_args, 0, &collect_error);
-                if (collect_error) {
+                if (!collect_error) {
+                        gupnp_service_notify_value (service, var_name, &value);
+
+                        g_value_unset (&value);
+
+                } else {
                         g_warning ("Error collecting value: %s\n",
                                    collect_error);
 
                         g_free (collect_error);
-
-                        g_value_unset (&value);
-
-                        return;
                 }
 
-                gupnp_service_notify_value (service, var_name, &value);
-
-                g_value_unset (&value);
-
                 var_name = va_arg (var_args, const char *);
         }
 }
@@ -1418,8 +1402,16 @@ notify_got_response (SoupMessage *msg,
 {
         SubscriptionData *data;
 
+        /* Cancelled? */
+        if (msg->status_code == SOUP_STATUS_CANCELLED)
+                return;
+
         data = user_data;
 
+        /* Remove from pending messages list */
+        data->service->priv->pending_messages = 
+                g_list_remove (data->service->priv->pending_messages, msg);
+
         if (SOUP_STATUS_IS_SUCCESSFUL (msg->status_code)) {
                 /* Success: reset callbacks pointer */
                 data->callbacks = g_list_first (data->callbacks);
@@ -1438,6 +1430,11 @@ notify_got_response (SoupMessage *msg,
                         soup_uri_free (uri);
 
                         /* And re-queue */
+                        data->service->priv->pending_messages = 
+                                g_list_prepend
+                                        (data->service->priv->pending_messages,
+                                         msg);
+
                         context = gupnp_service_info_get_context
                                         (GUPNP_SERVICE_INFO (data->service));
                         session = _gupnp_context_get_session (context);
@@ -1449,7 +1446,7 @@ notify_got_response (SoupMessage *msg,
 
                         error = g_error_new (GUPNP_EVENTING_ERROR,
                                              GUPNP_EVENTING_ERROR_NOTIFY_FAILED,
-                                             "Notify failed");
+                                             msg->reason_phrase);
 
                         g_signal_emit (data->service,
                                        signals[NOTIFY_FAILED],
@@ -1525,9 +1522,13 @@ notify_subscriber (gpointer key,
         msg->request.length = strlen (property_set);
 
         /* Queue */
+        data->service->priv->pending_messages = 
+                g_list_prepend (data->service->priv->pending_messages, msg);
+
         context = gupnp_service_info_get_context
                         (GUPNP_SERVICE_INFO (data->service));
         session = _gupnp_context_get_session (context);
+
         soup_session_queue_message (session,
                                     msg,
                                     notify_got_response,
@@ -1535,73 +1536,44 @@ notify_subscriber (gpointer key,
 }
 
 /* Create a property set from @queue */
-static xmlChar *
+static char *
 create_property_set (GQueue *queue)
 {
-        GValue transformed_value = {0, };
         NotifyData *data;
-        xmlDoc *doc;
-        xmlNode *node;
-        xmlNs *ns;
-        xmlChar *mem;
-        int size;
+        GString *str;
 
         /* Compose property set */
-        doc = xmlNewDoc ((const xmlChar *) "1.0");
+        str = xml_util_new_string ();
 
-        node = xmlNewDocNode (doc,
-                              NULL,
-                              (const xmlChar *) "propertyset",
-                              NULL);
-        xmlDocSetRootElement (doc, node);
-
-        ns = xmlNewNs (node,
-                       (const xmlChar *) "urn:schemas-upnp-org:event-1-0",
-                       (const xmlChar *) "e");
-        xmlSetNs (node, ns);
-
-        node = xmlNewTextChild (node,
-                                ns,
-                                (const xmlChar *) "property",
-                                NULL);
+        g_string_append (str,
+                         "<?xml version=\"1.0\"?>"
+                         "<e:propertyset xmlns:e="
+                                "\"urn:schemas-upnp-org:event-1-0\">"
+                         "<e:property>");
 
         /* Add variables */
         while ((data = g_queue_pop_head (queue))) {
-                g_value_init (&transformed_value, G_TYPE_STRING);
-                if (g_value_transform (&data->value, &transformed_value)) {
-                        /* Add to property set */
-                        xmlNewTextChild
-                                (node,
-                                 NULL,
-                                 (const xmlChar *) data->variable,
-                                 (xmlChar *) g_value_get_string
-                                                    (&transformed_value));
-                } else {
-                        g_warning ("Failed to transform value of type %s to "
-                                   "string.", G_VALUE_TYPE_NAME (&data->value));
-                }
+                xml_util_start_element (str, data->variable);
+                gvalue_util_value_append_to_xml_string (&data->value, str);
+                xml_util_end_element (str, data->variable);
 
                 /* Cleanup */
-                g_value_unset (&transformed_value);
-
                 notify_data_free (data);
         }
 
-        /* Dump document */
-        xmlDocDumpMemory (doc, &mem, &size);
+        g_string_append (str,
+                         "</e:property>"
+                         "</e:propertyset>");
 
-        /* Cleanup */
-        xmlFreeDoc (doc);
-
-        /* Return */
-        return mem;
+        /* Cleanup & return */
+        return g_string_free (str, FALSE);
 }
 
 /* Flush all queued notifications */
 static void
 flush_notifications (GUPnPService *service)
 {
-        xmlChar *mem;
+        char *mem;
 
         /* Create property set */
         mem = create_property_set (service->priv->notify_queue);
@@ -1612,7 +1584,7 @@ flush_notifications (GUPnPService *service)
                               mem);
 
         /* Cleanup */
-        xmlFree (mem);
+        g_free (mem);
 }
 
 /**
@@ -1683,41 +1655,182 @@ gupnp_service_thaw_notify (GUPnPService *service)
         flush_notifications (service);
 }
 
-/**
- * _gupnp_service_new
- * @context: A #GUPnPContext
- * @root_device: The #GUPnPRootDevice
- * @element: The #xmlNode ponting to the right service element
- * @udn: The UDN of the device the service is contained in
- * @location: The location of the service description file
- * @url_base: The URL base for this service
- *
- * Return value: A #GUPnPService for the service with element @element, as
- * read from the service description file specified by @location.
- **/
-GUPnPService *
-_gupnp_service_new (GUPnPContext    *context,
-                    GUPnPRootDevice *root_device,
-                    xmlNode         *element,
-                    const char      *udn,
-                    const char      *location,
-                    const SoupUri   *url_base)
+/* Convert a CamelCase string to a lowercase string with underscores */
+static char *
+strip_camel_case (char *camel_str)
 {
-        GUPnPService *service;
+        char *stripped;
+        int i, j;
 
-        g_return_val_if_fail (GUPNP_IS_CONTEXT (context), NULL);
-        g_return_val_if_fail (element != NULL, NULL);
-        g_return_val_if_fail (location != NULL, NULL);
-        g_return_val_if_fail (url_base != NULL, NULL);
+        /* Keep enough space for underscores */
+        stripped = g_malloc (strlen (camel_str) * 2);
 
-        service = g_object_new (GUPNP_TYPE_SERVICE,
-                                "context", context,
-                                "root-device", root_device,
-                                "location", location,
-                                "udn", udn,
-                                "url-base", url_base,
-                                "element", element,
-                                NULL);
+        for (i = 0, j = 0; i <= strlen (camel_str); i++) {
+                /* Convert every upper case letter to lower case and unless
+                 * it's the first character, the last charachter, in the
+                 * middle of an abbreviation or there is already an underscore
+                 * before it, add an underscore before it */
+                if (g_ascii_isupper (camel_str[i])) {
+                        if (i != 0 &&
+                            camel_str[i + 1] != '\0' &&
+                            camel_str[i - 1] != '_' &&
+                            !g_ascii_isupper (camel_str[i - 1])) {
+                                stripped[j++] = '_';
+                        }
+                        stripped[j++] = g_ascii_tolower (camel_str[i]);
+                } else
+                        stripped[j++] = camel_str[i];
+        }
 
-        return service;
+        return stripped;
+}
+
+/* Use the strings from @name_list as details to @signal_name, and connect
+ * callbacks with names based on these same strings to @signal_name::string. */
+static void
+connect_names_to_signal_handlers (GUPnPService *service,
+                                  GModule      *module,
+                                  const GList  *name_list,
+                                  const char   *signal_name,
+                                  const char   *callback_prefix,
+                                  gpointer      user_data)
+{
+        const GList *name_node;
+
+        for (name_node = name_list;
+             name_node;
+             name_node = name_node->next) {
+                GCallback callback;
+                char     *callback_name;
+                char     *signal_detail;
+                char     *tmp;
+
+                signal_detail = (char *) name_node->data;
+                tmp = strip_camel_case (signal_detail);
+
+                if (callback_prefix) {
+                        callback_name = g_strjoin ("_",
+                                                   callback_prefix,
+                                                   tmp,
+                                                   NULL);
+                        g_free (tmp);
+                        tmp = callback_name;
+                }
+
+                /* First try with 'on_' prefix */
+                callback_name = g_strjoin ("_",
+                                           "on",
+                                           tmp,
+                                           NULL);
+                if (!g_module_symbol (module,
+                                      callback_name,
+                                      (gpointer) &callback)) {
+                        g_free (callback_name);
+                        /* Now try with '_cb' postfix */
+                        callback_name = g_strjoin ("_",
+                                                   tmp,
+                                                   "cb",
+                                                   NULL);
+                        if (!g_module_symbol (module,
+                                              callback_name,
+                                              (gpointer) &callback)) {
+                                g_free (tmp);
+                                g_free (callback_name);
+
+                                continue;
+                        }
+                }
+
+                g_free (tmp);
+                g_free (callback_name);
+
+                signal_detail = g_strjoin ("::",
+                                           signal_name,
+                                           signal_detail,
+                                           NULL);
+
+                g_signal_connect (service,
+                                  signal_detail,
+                                  callback,
+                                  user_data);
+
+                g_free (signal_detail);
+        }
+}
+
+/**
+ * gupnp_service_signals_autoconnect
+ * @service: A #GUPnPService
+ * @user_data: the data to pass to each of the callbacks
+ * @error: return location for a GError, or NULL
+ *
+ * A convenience function that attempts to connect all possible "action-invoked"
+ * and "query-variable" signals to appropriate callbacks for the service
+ * @service. It uses service introspection and GModule's introspective
+ * features. It is very simillar to glade_xml_signal_autoconnect except that it
+ * attempts to guess the names of the signal handlers on its own.
+ *
+ * For this function to do its magic, the application must name the callback
+ * functions for "action-invoked" signals by striping the CamelCase off the
+ * action names and either prepend "on_" or append "_cb" to them. Same goes
+ * for "query-variable" signals, except that "query_" should be prepended to
+ * the variable name. For example, callback function for "GetSystemUpdateID"
+ * action should be either named as "get_system_update_id_cb" or
+ * "on_get_system_update_id" and callback function for the query of
+ * "SystemUpdateID" state variable should be named "query_system_update_id_cb"
+ * or "on_query_system_update_id".
+ *
+ * Note that this function will not work correctly if GModule is not supported
+ * on the platform or introspection is not available for service @service.
+ *
+ * WARNING: This function can not and therefore does not guarantee that the
+ * resulting signal connections will be correct as it depends heavily on a
+ * particular naming schemes described above.
+ **/
+void
+gupnp_service_signals_autoconnect (GUPnPService *service,
+                                   gpointer      user_data,
+                                   GError      **error)
+{
+        GUPnPServiceIntrospection *introspection;
+        const GList               *names;
+        GModule                   *module;
+
+        g_return_if_fail (GUPNP_IS_SERVICE (service));
+
+        introspection = gupnp_service_info_get_introspection
+                                (GUPNP_SERVICE_INFO (service),
+                                 error);
+        if (!introspection)
+                return;
+
+        /* Get a handle on the main executable -- use this to find symbols */
+        module = g_module_open (NULL, 0);
+        if (module == NULL) {
+                g_error ("Failed to open module: %s", g_module_error ());
+
+                g_object_unref (introspection);
+
+                return;
+        }
+
+        names = gupnp_service_introspection_list_action_names (introspection);
+        connect_names_to_signal_handlers (service,
+                                          module,
+                                          names,
+                                          "action-invoked",
+                                          NULL,
+                                          user_data);
+
+        names = gupnp_service_introspection_list_state_variable_names
+                        (introspection);
+        connect_names_to_signal_handlers (service,
+                                          module,
+                                          names,
+                                          "query-variable",
+                                          "query",
+                                          user_data);
+
+        g_module_close (module);
+        g_object_unref (introspection);
 }
